(*----------------------------------------------------------------------------*
 *  Direct3D Shader Disassembler                                              *
 *                                                                            *
 *  "ShaderLexer" unit contains shader disassembler with embedded simple way  *
 *  to export disassembled data as text output. But actually it was designed  *
 *  to be used for shader data analis. So other usage scenarios possible      *
 *                                                                            *
 *  Copyright (c) 2005 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@clootie.ru                                                *
 *                                                                            *
 *  Modified: 05-May-2005                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://www.clootie.ru                                                  *
 *----------------------------------------------------------------------------*)
 
unit ShaderLexer;

interface

uses
  Windows, Classes, SysUtils, Direct3D9, D3DX9;

type
  TD3DTokenType = (ttVersion, ttComment, ttInstruction, ttParameter, ttEnd);

  TD3DShaderState = (ssVersion, ssHeader{Declaration/Definition}, ssBody, ssInvalidStream);

  TD3DShaderLexer = class
  private
    FInstructionLength: DWORD;
    FInstructionStartIndex: DWORD;
    FInstructionCode: DWORD;
    FInstructionAsString: String;
    fEOF: Boolean;
    fExtendedPrecisionContants: Boolean;
    function GetData(Index: DWORD): DWORD;
    function GetDeclarationUsage2(Token: DWORD): String;
  private
    FFileName: String;
    FBuffer: PDWORD;  // Shader code
    FPointer: PDWORD; // Pointer to currently analizing code
    FBufferSize: Integer;
  private
    FShaderVersion: DWORD;
    FState: TD3DShaderState;
    // parameters - params
    function GetToken: DWORD;
    function GetRegisterName(Token: DWORD): String;
    function GetDeclarationUsage(Token: DWORD): String;
  protected
    function GetTokenType(Token: DWORD): TD3DTokenType;
    function GetVersionText(Token: DWORD): String;
    function GetCommentText(CommentToken: DWORD): String;
    function GetInstructionText(Token: DWORD): String;
    function GetDestinParamText: String;
    function GetSourceParamText: String;
    function AddRegisterToCommand(IsFirstRegister: Boolean; Instr, Reg: String): String;
    ////////////////////////////////////////////////////////////
    property Token: DWORD read GetToken;
  public
    constructor Create(aFileName: String);
    constructor CreateFromStream(S: TStream);
    destructor Destroy; override;
    ////////////////////////////////////////////////////////////
    procedure Reset;
    function NextInstruction: Boolean;
    function IsVertexShader: Boolean;
    ////////////////////////////////////////////////////////////
    procedure SaveToFile(FileName: String);
    procedure SaveToStream(S: TStream); virtual;
    function VerifyDisassembly(out Correct: Boolean; out Errors: String): Boolean;
    ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
    property ExtendedPrecisionContants: Boolean read fExtendedPrecisionContants write fExtendedPrecisionContants;
    property ShaderVersion: DWORD read FShaderVersion;
    property State: TD3DShaderState read FState;
    ////////////////////////////////////////////////////////////
    property InstructionCode: DWORD read FInstructionCode;
    property InstructionAsString: String read FInstructionAsString;
    property InstructionStartIndex: DWORD read FInstructionStartIndex;
    property InstructionLength: DWORD read FInstructionLength; // Length of instruction in DWORD's
    property EOF: Boolean read fEOF;
    property Data[Index: DWORD]: DWORD read GetData;
  end;

  EShaderException = Exception;

implementation


{ TD3DShaderLexer }

constructor TD3DShaderLexer.Create(aFileName: String);
var
  S: TFileStream;
begin
  FFileName:= aFileName;

  S:= TFileStream.Create(FFileName, fmOpenRead);
  try
    CreateFromStream(S);
  finally
    S.Free;
  end;
end;

constructor TD3DShaderLexer.CreateFromStream(S: TStream);
begin
  FBufferSize:= S.Size;
  GetMem(FBuffer, FBufferSize);
  S.Read(FBuffer^, FBufferSize);

  Reset;
end;

destructor TD3DShaderLexer.Destroy;
begin
  FreeMem(FBuffer);
  inherited;
end;

function TD3DShaderLexer.GetToken: DWORD;
begin
  Result:= FPointer^;
end;

function TD3DShaderLexer.GetVersionText(Token: DWORD): String;
begin
  case (Token shr 16) of
    $FFFF: Result:= 'ps';
    $FFFE: Result:= 'vs';
   else
     raise EShaderException.Create('Version Token');
  end;

  if (D3DSHADER_VERSION_MINOR(Token) = $FF) then
    Result:= Format('%s_%d_sw', [Result, D3DSHADER_VERSION_MAJOR(Token)])
  else
  if (D3DSHADER_VERSION_MINOR(Token) = $01) and (D3DSHADER_VERSION_MAJOR(Token) > $01) then
    Result:= Format('%s_%d_x',  [Result, D3DSHADER_VERSION_MAJOR(Token)])
  else
    Result:= Format('%s_%d_%d', [Result, D3DSHADER_VERSION_MAJOR(Token),  D3DSHADER_VERSION_MINOR(Token)]);

  Result:= '    ' + Result;
end;

function TD3DShaderLexer.GetCommentText(CommentToken: DWORD): String;
var
  CommentSize: DWORD;
begin
  CommentSize:= (CommentToken and D3DSI_COMMENTSIZE_MASK) shr D3DSI_COMMENTSIZE_SHIFT;
  FInstructionLength:= CommentSize + 1;
  Result:= '; Debug info embedded here'#10;  
  Inc(FPointer, CommentSize);
end;

type
  TInstrCode = record
    code: DWORD;
    text: PChar;
    params: DWORD;
  end;

const
  Codes: array[0..84] of TInstrCode = (
    (code: D3DSIO_NOP;          text: 'nop';          params: 0),
    (code: D3DSIO_MOV;          text: 'mov';          params: 2),
    (code: D3DSIO_ADD;          text: 'add';          params: 3),
    (code: D3DSIO_SUB;          text: 'sub';          params: 3),
    (code: D3DSIO_MAD;          text: 'mad';          params: 4),
    (code: D3DSIO_MUL;          text: 'mul';          params: 3),
    (code: D3DSIO_RCP;          text: 'rcp';          params: 2),
    (code: D3DSIO_RSQ;          text: 'rsq';          params: 2),
    (code: D3DSIO_DP3;          text: 'dp3';          params: 3),
    (code: D3DSIO_DP4;          text: 'dp4';          params: 3),
    (code: D3DSIO_MIN;          text: 'min';          params: 3),
    (code: D3DSIO_MAX;          text: 'max';          params: 3),
    (code: D3DSIO_SLT;          text: 'slt';          params: 3),
    (code: D3DSIO_SGE;          text: 'sge';          params: 3),
    (code: D3DSIO_EXP;          text: 'exp';          params: $FFFFFFFF),
    (code: D3DSIO_LOG;          text: 'log';          params: $FFFFFFFF),
    (code: D3DSIO_LIT;          text: 'lit';          params: $FFFFFFFF),
    (code: D3DSIO_DST;          text: 'dst';          params: $FFFFFFFF),
    (code: D3DSIO_LRP;          text: 'lrp';          params: $FFFFFFFF),
    (code: D3DSIO_FRC;          text: 'frc';          params: 2),
    (code: D3DSIO_M4x4;         text: 'm4x4';         params: $FFFFFFFF),
    (code: D3DSIO_M4x3;         text: 'm4x3';         params: $FFFFFFFF),
    (code: D3DSIO_M3x4;         text: 'm3x4';         params: $FFFFFFFF),
    (code: D3DSIO_M3x3;         text: 'm3x3';         params: $FFFFFFFF),
    (code: D3DSIO_M3x2;         text: 'm3x2';         params: $3),
    (code: D3DSIO_CALL;         text: 'call';         params: $FFFFFFFF),
    (code: D3DSIO_CALLNZ;       text: 'callnz';       params: $FFFFFFFF),
    (code: D3DSIO_LOOP;         text: 'loop';         params: $FFFFFFFF),
    (code: D3DSIO_RET;          text: 'ret';          params: $0),
    (code: D3DSIO_ENDLOOP;      text: 'endloop';      params: $FFFFFFFF),
    (code: D3DSIO_LABEL;        text: 'label ';       params: $FFFFFFFF),
    (code: D3DSIO_DCL;          text: 'dcl';          params: $2),
    (code: D3DSIO_POW;          text: 'pow';          params: $FFFFFFFF),
    (code: D3DSIO_CRS;          text: 'crs';          params: $FFFFFFFF),
    (code: D3DSIO_SGN;          text: 'sgn';          params: $FFFFFFFF),
    (code: D3DSIO_ABS;          text: 'abs';          params: $FFFFFFFF),
    (code: D3DSIO_NRM;          text: 'nrm';          params: $FFFFFFFF),
    (code: D3DSIO_SINCOS;       text: 'sincos';       params: $FFFFFFFF),
    (code: D3DSIO_REP;          text: 'rep';          params: $FFFFFFFF),
    (code: D3DSIO_ENDREP;       text: 'endrep';       params: $FFFFFFFF),
    (code: D3DSIO_IF;           text: 'if';           params: $FFFFFFFF),
    (code: D3DSIO_IFC;          text: 'if';           params: $FFFFFFFF),
    (code: D3DSIO_ELSE;         text: 'else';         params: $FFFFFFFF),
    (code: D3DSIO_ENDIF;        text: 'endif';        params: $FFFFFFFF),
    (code: D3DSIO_BREAK;        text: 'break';        params: $FFFFFFFF),
    (code: D3DSIO_BREAKC;       text: 'break';        params: $FFFFFFFF),
    (code: D3DSIO_MOVA;         text: 'mova';         params: $FFFFFFFF),
    (code: D3DSIO_DEFB;         text: 'defb';         params: $2),
    (code: D3DSIO_DEFI;         text: 'defi';         params: $5),
                                                                          // checked till here !!!
    (code: D3DSIO_TEXCOORD;     text: 'texcoord';     params: $FFFFFFFF),
    (code: D3DSIO_TEXKILL;      text: 'texkill ';     params: $FFFFFFFF),
    (code: D3DSIO_TEX;          text: 'tex';          params: $FFFFFFFF),
    (code: D3DSIO_TEXBEM;       text: 'texbem';       params: $FFFFFFFF),
    (code: D3DSIO_TEXBEML;      text: 'texbeml ';     params: $FFFFFFFF),
    (code: D3DSIO_TEXREG2AR;    text: 'texreg2ar';    params: $FFFFFFFF),
    (code: D3DSIO_TEXREG2GB;    text: 'texreg2gb';    params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x2PAD;   text: 'texm3x2pad';   params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x2TEX;   text: 'texm3x2tex';   params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x3PAD;   text: 'texm3x3pad';   params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x3TEX;   text: 'texm3x3tex';   params: $FFFFFFFF),
    (code: D3DSIO_RESERVED0;    text: 'reserved0';    params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x3SPEC;  text: 'texm3x3spec '; params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x3VSPEC; text: 'texm3x3vspec'; params: $FFFFFFFF),
    (code: D3DSIO_EXPP;         text: 'expp';         params: $FFFFFFFF),
    (code: D3DSIO_LOGP;         text: 'logp';         params: $FFFFFFFF),
    (code: D3DSIO_CND;          text: 'cnd';          params: $FFFFFFFF),
    (code: D3DSIO_DEF;          text: 'def';          params: $5),
    (code: D3DSIO_TEXREG2RGB;   text: 'texreg2rgb';   params: $FFFFFFFF),
    (code: D3DSIO_TEXDP3TEX;    text: 'texdp3tex';    params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x2DEPTH; text: 'texm3x2depth'; params: $FFFFFFFF),
    (code: D3DSIO_TEXDP3;       text: 'texdp3';       params: $FFFFFFFF),
    (code: D3DSIO_TEXM3x3;      text: 'texm3x3';      params: $FFFFFFFF),
    (code: D3DSIO_TEXDEPTH;     text: 'texdepth';     params: $FFFFFFFF),
    (code: D3DSIO_CMP;          text: 'cmp';          params: $FFFFFFFF),
    (code: D3DSIO_BEM;          text: 'bem';          params: $FFFFFFFF),

    (code: D3DSIO_DP2ADD;       text: 'dp2add';       params: $FFFFFFFF),
    (code: D3DSIO_DSX;          text: 'dsx';          params: $FFFFFFFF),
    (code: D3DSIO_DSY;          text: 'dsy';          params: $FFFFFFFF),
    (code: D3DSIO_TEXLDD;       text: 'texldd ';      params: $FFFFFFFF),
    (code: D3DSIO_SETP;         text: 'setp';         params: $FFFFFFFF),
    (code: D3DSIO_TEXLDL;       text: 'texldl ';      params: $FFFFFFFF),
    (code: D3DSIO_BREAKP;       text: 'breakp';       params: $FFFFFFFF),


    (code: D3DSIO_PHASE;        text: 'phase';        params: $FFFFFFFF),
    (code: D3DSIO_COMMENT;      text: 'comment';      params: $FFFFFFFF),
    (code: D3DSIO_END;          text: 'end';          params: $FFFFFFFF)
  );

function GetInstructionModifier(Token: DWORD): String;
const
  Comparisions: array [D3DSPC_GT..D3DSPC_LE] of PChar =
    ('gt', 'eq', 'ge', 'lt', 'ne', 'le');
var
 Comp: TD3DShaderComparison;
begin
  Comp:= TD3DShaderComparison(
         (Token and D3DSHADER_COMPARISON_MASK) shr D3DSHADER_COMPARISON_SHIFT);
  if not (Comp in [D3DSPC_RESERVED0, D3DSPC_RESERVED1]) then
    Result:= '_' + Comparisions[Comp]
  else
    Result:= '';
end;

function GetInstructionModifier2(Token: DWORD): String;
begin
  Result:= '';

  if (Token and D3DSP_DSTSHIFT_MASK <> 0) then
  begin
    case ((Token and D3DSP_DSTSHIFT_MASK) shr D3DSP_DSTSHIFT_SHIFT) of
      // These values are not defined in DirectX SDK headers
      $1: Result:= Result + '_x2';
      $2: Result:= Result + '_x4';
      $3: Result:= Result + '_x8';
      $D: Result:= Result + '_d8';
      $E: Result:= Result + '_d4';
      $F: Result:= Result + '_d2';
    end;
  end;

  if (Token and D3DSP_DSTMOD_MASK <> 0) then
  begin
    if (Token and D3DSP_DSTMOD_MASK = D3DSPDM_SATURATE)
    then Result:= Result + '_sat';
    if (Token and D3DSP_DSTMOD_MASK = D3DSPDM_PARTIALPRECISION)
    then Result:= Result + '_pp';
    if (Token and D3DSP_DSTMOD_MASK = D3DSPDM_MSAMPCENTROID)
    then Result:= Result + '_centroid';
  end;

  if Result <> '' then Result:= Result + ' ';
end;

function TD3DShaderLexer.GetInstructionText(Token: DWORD): String;
var
  i: Integer;
  t: DWORD;
begin
  Result:= '';
  t:= Token;

  for i:= 0 to High(Codes) do
  begin
    if Codes[i].code = (t and D3DSI_OPCODE_MASK) then
    begin
      Result:= Codes[i].text;
      FInstructionCode:= Codes[i].Code;
      Break;
    end;
  end;

(*
The tex and texcoord instructions apply to pixel shader versions 1.0 through 1.3;
    each instruction has one destination parameter associated with it.

The texld and texcrd instructions are new to pixel shader version 1.4 and later;
    each instruction has both destination and source parameters associated with it.

The runtime converts the tex and texld user-mode instructions to the D3DSIO_TEX kernel-mode operation.
The runtime converts the texcoord and texcrd user-mode instructions to the D3DSIO_TEXCOORD kernel-mode operation.
Drivers first verify the pixel shader version of the shader code and then process the instruction accordingly.
For example, if a driver verifies that it received version 1.4 pixel shader code with a D3DSIO_TEX
operation, the driver determines that destination and source parameters follow the instruction token.
*)
  if (t and D3DSI_OPCODE_MASK) = D3DSIO_TEX then
  begin
    if (D3DSHADER_VERSION_MAJOR(ShaderVersion)>1) or 
       (D3DSHADER_VERSION_MINOR(ShaderVersion)>3)
      then Result:= 'texld' // For PS > 1.3
      else Result:= 'tex';
  end;
  if (t and D3DSI_OPCODE_MASK) = D3DSIO_TEXCOORD then
  begin
    if (D3DSHADER_VERSION_MAJOR(ShaderVersion)=1) and
       (D3DSHADER_VERSION_MINOR(ShaderVersion)>3)
      then Result:= 'texcoord'
      else Result:= 'texcrd';
  end;

  if (t and D3DSI_OPCODE_MASK) in [D3DSIO_BREAKC, D3DSIO_IFC, D3DSIO_SETP] then
    Result:= Result + GetInstructionModifier(t);

  if (Token and D3DSI_COISSUE) = D3DSI_COISSUE
  then Result:= '  + ' + Result
  else Result:= '    ' + Result;
end;

function TD3DShaderLexer.GetDeclarationUsage(Token: DWORD): String;
const
  DeclUsageStrings: array[D3DDECLUSAGE_POSITION..D3DDECLUSAGE_SAMPLE] of PChar =
  (
    'position',
    'blendweight',
    'blendindices',
    'normal',
    'psize',
    'texcoord',
    'tangent',
    'binormal',
    'tessfactor',
    'positiont',
    'color',
    'fog',
    'depth',
    'sample'
  );
begin
  Result:=
    Format('_%s%d', [
      DeclUsageStrings[TD3DDeclUsage((Token and D3DSP_DCL_USAGE_MASK) shr D3DSP_DCL_USAGE_SHIFT)],
      (Token and D3DSP_DCL_USAGEINDEX_MASK) shr D3DSP_DCL_USAGEINDEX_SHIFT
    ]);
  if Result[Length(Result)] = '0' then SetLength(Result, Length(Result)-1);
end;

function TD3DShaderLexer.GetDeclarationUsage2(Token: DWORD): String;
begin
  Result:= '';
  case TD3DSamplerTextureType(
         (Token and D3DSP_TEXTURETYPE_MASK)) of
    D3DSTT_2D: Result:= '2d';
    D3DSTT_CUBE: Result:= 'cube';
    D3DSTT_VOLUME: Result:= 'volume';
  end;
  if Result <> '' then Result:= '_' + Result;
end;

function TD3DShaderLexer.GetRegisterName(Token: DWORD): String;
const
  RegisterStrings: array[D3DSPR_TEMP..D3DSPR_PREDICATE] of PChar =
  (
    'r',    // D3DSPR_TEMP        =  0, // Temporary Register File
    'v',    // D3DSPR_INPUT       =  1, // Input Register File
    'c',    // D3DSPR_CONST       =  2, // Constant Register File
    '', {+} // D3DSPR_ADDR        =  3, // Address Register (VS)
//    '',    // D3DSPR_TEXTURE     =  3, // Texture Register File (PS)
    '', {+} // D3DSPR_RASTOUT     =  4, // Rasterizer Register File
    'oD',   // D3DSPR_ATTROUT     =  5, // Attribute Output Register File
    '', {+} // D3DSPR_TEXCRDOUT   =  6, // Texture Coordinate Output Register File
//    '',     // D3DSPR_OUTPUT      =  6, // Output register file for VS3.0+
    'i',    // D3DSPR_CONSTINT    =  7, // Constant Integer Vector Register File
    'oC',   // D3DSPR_COLOROUT    =  8, // Color Output Register File
    'oDepth', // D3DSPR_DEPTHOUT    =  9, // Depth Output Register File
    's',    // D3DSPR_SAMPLER     = 10, // Sampler State Register File
    'c',    // D3DSPR_CONST2      = 11, // Constant Register File  2048 - 4095
    'c',    // D3DSPR_CONST3      = 12, // Constant Register File  4096 - 6143
    'c',    // D3DSPR_CONST4      = 13, // Constant Register File  6144 - 8191
    'b',    // D3DSPR_CONSTBOOL   = 14, // Constant Boolean register file
    'aL',{+}// D3DSPR_LOOP        = 15, // Loop counter register file
    '', {-} // D3DSPR_TEMPFLOAT16 = 16, // 16-bit float temp register file
    '', {+} // D3DSPR_MISCTYPE    = 17, // Miscellaneous (single) registers.
    'l',    // D3DSPR_LABEL       = 18, // Label
    'p'     // D3DSPR_PREDICATE   = 19  // Predicate register
  );
var
  RegisterType: TD3DShaderParamRegisterType;
  ApppendNumber: Boolean;
  RegNumber: Integer;
begin
  ApppendNumber:= True;
  RegisterType:= TD3DShaderParamRegisterType(
        // In DirectX9 Register type in encoded in two bit fields:
        ((Token and D3DSP_REGTYPE_MASK ) shr D3DSP_REGTYPE_SHIFT ) or
        ((Token and D3DSP_REGTYPE_MASK2) shr D3DSP_REGTYPE_SHIFT2));
  Result:= RegisterStrings[RegisterType];

  case RegisterType of
    D3DSPR_ADDR: // == D3DSPR_TEXTURE
      if IsVertexShader then Result:= 'a' else Result:= 't';
    D3DSPR_TEXCRDOUT:
      if IsVertexShader and (D3DSHADER_VERSION_MAJOR(ShaderVersion)>=3)
        then Result:= 'o'
        else Result:= 'oT';
    D3DSPR_RASTOUT:
    begin
      case TD3DVSRastoutOffsets(Token and D3DSP_REGNUM_MASK) of
        D3DSRO_POSITION:   Result:= 'oPos';
        D3DSRO_FOG:        Result:= 'oFog';
        D3DSRO_POINT_SIZE: Result:= 'oPts';
      end;
      ApppendNumber:= False;
    end;
    D3DSPR_MISCTYPE:
    begin
      case TD3DShaderMiscTypeOffsets(Token and D3DSP_REGNUM_MASK) of
        D3DSMO_POSITION: Result:= 'vPos';
        D3DSMO_FACE:     Result:= 'vFace';
      end;
      ApppendNumber:= False;
    end;
    D3DSPR_LOOP:
    begin
      Result:= 'aL';
      ApppendNumber:= False;
    end;
    D3DSPR_DEPTHOUT:
      ApppendNumber:= False;
   else
  end;
  if Result = '' then
    Result:= Format('unknown[%d]', [Integer(RegisterType)]);

  RegNumber:= (Token and D3DSP_REGNUM_MASK);
  if RegisterType in [D3DSPR_CONST2, D3DSPR_CONST3, D3DSPR_CONST4] then
  begin
    case RegisterType of
      D3DSPR_CONST2: Inc(RegNumber, 2048);
      D3DSPR_CONST3: Inc(RegNumber, 4096);
      D3DSPR_CONST4: Inc(RegNumber, 6144);
    end;
  end;
  
  if ApppendNumber then
    Result:= Result + IntToStr(RegNumber);
end;

function GetWriteMaskText(Token: DWORD): String;
begin
  if Token and D3DSP_WRITEMASK_ALL <> D3DSP_WRITEMASK_ALL then
  begin
    Result:= '.';
    if (Token and D3DSP_WRITEMASK_0 = D3DSP_WRITEMASK_0) then Result:= Result + 'x';
    if (Token and D3DSP_WRITEMASK_1 = D3DSP_WRITEMASK_1) then Result:= Result + 'y';
    if (Token and D3DSP_WRITEMASK_2 = D3DSP_WRITEMASK_2) then Result:= Result + 'z';
    if (Token and D3DSP_WRITEMASK_3 = D3DSP_WRITEMASK_3) then Result:= Result + 'w';
  end else
    Result:= '';
end;

function GetSourceSwizzleText(Token: DWORD): String;
var
  i: Integer;
begin
  if Token and D3DVS_SWIZZLE_MASK <> D3DVS_NOSWIZZLE then
  begin
    Token:= Token shr D3DVS_SWIZZLE_SHIFT;
    for i:= 1 to 4 do
    begin
      SetLength(Result, 4);
      case Token and 3 of
        0: Result[i]:= 'x';
        1: Result[i]:= 'y';
        2: Result[i]:= 'z';
        3: Result[i]:= 'w';
      end;
      Token:= Token shr 2;
    end;
    if (Result[1] = Result[2]) and (Result[3] = Result[4]) and (Result[2] = Result[3]) then Result:= Result[1];
    Result:= '.' + Result;
  end else
    Result:= '';
end;


function TD3DShaderLexer.GetDestinParamText: String;
begin
  Result:= GetInstructionModifier2(Token) + GetRegisterName(Token) + GetWriteMaskText(Token);
end;


function ModifySourceParam(Token: DWORD; Param: String): String;
begin
  if Token and D3DSP_SRCMOD_MASK <> 0 then
  begin
    case (Token and D3DSP_SRCMOD_MASK) of
    //D3DSPSM_NONE        = 0; // nop
      D3DSPSM_NEG        : Result:= '-' + Param; // negate
      D3DSPSM_BIAS       : Result:=       Param + '_bias'; // bias
      D3DSPSM_BIASNEG    : Result:= '-' + Param + '_bias'; // bias and negate
      D3DSPSM_SIGN       : Result:=       Param + '_bx2'; // sign
      D3DSPSM_SIGNNEG    : Result:= '-' + Param + '_bx2'; // sign and negate
      D3DSPSM_COMP       : Result:= '1-'+ Param; // complement
      D3DSPSM_X2         : Result:=       Param + '_x2'; // *2
      D3DSPSM_X2NEG      : Result:= '-' + Param + '_x2'; // *2 and negate
      D3DSPSM_DZ         : Result:=       Param + '_dz'; // divide through by z component
      D3DSPSM_DW         : Result:=       Param + '_dw'; // divide through by w component
      D3DSPSM_ABS        : Result:=       Param + '_abs'; // abs()
      D3DSPSM_ABSNEG     : Result:= '-' + Param + '_abs'; // -abs()
      D3DSPSM_NOT        : Result:= '!' + Param; // for predicate register: "!p0"
    end;
  end else
    Result:= Param;
end;

function TD3DShaderLexer.GetSourceParamText: String;
begin
  Result:= ModifySourceParam(Token, GetRegisterName(Token)) + GetSourceSwizzleText(Token);
end;

function TD3DShaderLexer.IsVertexShader: Boolean;
begin
  Result:= (FShaderVersion and DWORD($FFFF shl 16)) = DWORD($FFFE shl 16);
end;

function TD3DShaderLexer.GetTokenType(Token: DWORD): TD3DTokenType;
begin
  if (Token = $FFFF) then
  begin // END instruction
    Result:= ttEnd;
  end else
  if (Token and $FFFF = $FFFE) then
  begin // Comment Token
    Result:= ttComment;
  end else
  if (Token and DWORD(1 shl 31) = DWORD(1 shl 31)) then
  begin // either Destin or Source parameter
    Result:= ttParameter
  end else
  begin // Instruction code
    Result:= ttInstruction;
  end;
end;

function TD3DShaderLexer.GetData(Index: DWORD): DWORD;
begin
  Result:= PDWORDArray(FBuffer)[Index-1];
end;

function TD3DShaderLexer.NextInstruction;
 procedure ParseDefines;
 var
   S, S1: String;
   i, i_max: Integer;
   Code: DWORD;
   FmtSettings: TFormatSettings;
 begin
   GetLocaleFormatSettings(GetThreadLocale, FmtSettings);
   FmtSettings.DecimalSeparator:= '.';
   
   S:= FInstructionAsString + ' ' + GetDestinParamText;
   Inc(FPointer);

   case InstructionCode of
     D3DSIO_DEFB: i_max:= 1;
     D3DSIO_DEF,
     D3DSIO_DEFI: i_max:= 4;
    else
     raise EShaderException.Create('Wrong instruction code');
   end;

   FInstructionLength:= 2 + i_max;

   for i:= 1 to i_max do
   begin
     Code:= Token;
     case InstructionCode of
       D3DSIO_DEF:
         if ExtendedPrecisionContants then
           S1:= FormatFloat('0.0###################e-', PSingle(@Code)^, FmtSettings)
         else
           if Abs(PSingle(@Code)^) >= 1e6
           then S1:= FormatFloat('#.0e-', PSingle(@Code)^, FmtSettings)
           else S1:= FormatFloat('0.#####', PSingle(@Code)^, FmtSettings);
       D3DSIO_DEFI: S1:= Format('%d', [Code]);
       D3DSIO_DEFB: if Code = 1 then S1:= 'true' else S1:= 'false';
     end;
     S:= S + ', ' + S1;
     Inc(FPointer);
   end;
   FInstructionAsString:= S;
 end;

 procedure ParseDeclaration;
 var
   S: String;
 begin
   FInstructionLength:= 3;

   if IsVertexShader then S:= GetDeclarationUsage(Token) else S:= '';
   S:= S + GetDeclarationUsage2(Token);
   Inc(FPointer);
   FInstructionAsString:= AddRegisterToCommand(True, FInstructionAsString + S, GetDestinParamText);
   Inc(FPointer);
 end;

var
  SavedToken: DWORD;
  TokenType: TD3DTokenType;
  SingleParamInstruction: Boolean;
  TokenText: String;
begin
  // Check: Have we reached end of Shader?
  // (+3) is to overcome issue of wrong data (not compiled VS/PS) 
  fEOF:= (Integer(FPointer) - Integer(FBuffer) + 3 >= FBufferSize);
  Result:= not fEOF;
  if not Result then Exit;

  FInstructionStartIndex:= (DWORD(FPointer) - DWORD(FBuffer)) div SizeOf(DWORD) + 1;
  SavedToken:= Token;
  FInstructionCode:= (SavedToken and D3DSI_OPCODE_MASK);
  TokenType:= GetTokenType(SavedToken);
  FInstructionAsString:= GetInstructionText(SavedToken);
  Inc(FPointer);

  if State = ssVersion then
  begin
    FShaderVersion:= SavedToken;
    try
      FInstructionAsString:= GetVersionText(SavedToken);
    except
      FState:= ssInvalidStream;
      FInstructionLength:= FBufferSize div SizeOf(DWORD);
      FInstructionAsString:= '; Invalid stream data';
      Inc(FPointer, FInstructionLength-1);
      Exit;
    end;
    FInstructionLength:= 1;
    FState:= ssHeader; // moving state to next stage
  end else
  if State = ssHeader then
  begin // Either in Declaration or in Definition
    case (SavedToken and D3DSI_OPCODE_MASK) of
      D3DSIO_DEF,
      D3DSIO_DEFI,
      D3DSIO_DEFB: ParseDefines;      // "def" paramaters
      D3DSIO_DCL:  ParseDeclaration;  // "dcl_xxx" paramaters
      D3DSIO_COMMENT: FInstructionAsString:= GetCommentText(SavedToken); // Do nothing 
     else
       FState:= ssBody; // reached shader main body
    end;
  end;

  if State = ssBody then
  begin
    case TokenType of
      ttComment:
        FInstructionAsString:= GetCommentText(SavedToken);
      ttInstruction:
      begin
        FInstructionAsString:= GetInstructionText(SavedToken);
        FInstructionLength:= 1;
        while (Token and DWORD(1 shl 31) = DWORD(1 shl 31)) do // HIGH bit set => parameter token
        begin
          // Check if next token is instruction?
          // If so then where is only one parameter to instr => it's takes as source param
          Inc(FPointer);
          SingleParamInstruction:= (Token and DWORD(1 shl 31) = 0) and (InstructionLength = 1);
          Dec(FPointer);

          // One way around
          if InstructionCode in [D3DSIO_CALLNZ, D3DSIO_LOOP, D3DSIO_IFC, D3DSIO_BREAKP, D3DSIO_BREAKC]
          then TokenText:= GetSourceParamText
          else
          begin
            if SingleParamInstruction then
            begin
              if (InstructionCode = D3DSIO_TEX)
              then TokenText:= GetDestinParamText
              else TokenText:= GetSourceParamText;
            end else
            if InstructionLength = 1
              then TokenText:= GetDestinParamText
              else TokenText:= GetSourceParamText;
          end;

          FInstructionAsString:= AddRegisterToCommand((InstructionLength = 1),
                                    FInstructionAsString, TokenText);

          Inc(FPointer);
          Inc(FInstructionLength);
        end;
      end;
      ttEnd:
      begin
        FInstructionAsString:= '; end';
        FInstructionLength:= 1;
        Inc(FPointer);
      end;
    end;
  end;
  fEOF:= (Integer(FPointer) - Integer(FBuffer) >= FBufferSize);
end;

procedure TD3DShaderLexer.Reset;
begin
  FPointer:= FBuffer;
  FState:= ssVersion;
  NextInstruction;
end;

procedure TD3DShaderLexer.SaveToStream(S: TStream);
  procedure WriteString(Str: String);
  const
    CRLF: array[0..1] of Char = #13#10;
  begin
    S.Write(Str[1], Length(Str));
    S.Write(CRLF, 2);
  end;
var
  i: Integer;
  mm: string;
begin
  WriteString('//');
  WriteString('// Generated by Clootie DirectX9 Shader Disassembler');
  WriteString('//');
  WriteString('//  Latest version / documentation:');
  WriteString('//    http://www.clootie.ru/delphi/dxtools');
  WriteString('//    http://www.clootie.ru/articles/dxShaderDisassembler.html');
  WriteString('//');
  WriteString('');

  Reset;
  while not EOF do
  begin
    // WriteString(InstructionAsString);
    mm:= InstructionAsString + #9';';
    for i:= InstructionStartIndex to InstructionStartIndex + InstructionLength - 1 do
      mm:= mm + Format(' $%.8x', [Data[i]]);
    WriteString(mm);
    NextInstruction;
  end;
end;

procedure TD3DShaderLexer.SaveToFile(FileName: String);
var
  S: TFileStream;
begin
  S:= TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(S);
  finally
    S.Free;
  end;
end;

function TD3DShaderLexer.AddRegisterToCommand(IsFirstRegister: Boolean;
  Instr, Reg: String): String;
var
  Space: String;
begin
  if IsFirstRegister then Space:= ' ' else Space:= ', ';

  if (Length(Reg) > 0) and (Reg[1] = '_')
  then Result:= Instr + Reg
  else Result:= Instr + Space + Reg;
end;

type
  TD3DXAssembleShader = function (
    pSrcData: PAnsiChar;
    SrcDataLen: LongWord;
    pDefines: PD3DXMacro;
    pInclude: ID3DXInclude;
    Flags: DWord;
    ppShader: PID3DXBuffer;
    ppErrorMsgs: PID3DXBuffer): HResult; stdcall; // external d3dx9dll; 

function TD3DShaderLexer.VerifyDisassembly(out Correct: Boolean; out Errors: String): Boolean;
var
  Lib: THandle;
  D3DXAssembleShader: TD3DXAssembleShader;
  DAShader: TMemoryStream;
  ShaBuff, ErrBuff: ID3DXBuffer;
begin
  Result:= False; // Function call is FAILED by default

  Lib := 0;
  if (Lib = 0) then Lib:= LoadLibrary('d3dx9_25.dll'); // April 2005 SDK
  if (Lib = 0) then Lib:= LoadLibrary('d3dx9_24.dll'); // February 2005 SDK
  if (Lib = 0) then Lib:= LoadLibrary('D3DX92ab.dll'); // Summer 2003 - December 2004
  if (Lib = 0) then Lib:= LoadLibrary('D3DX9ab.dll');  // Before Summer 2003
  if (Lib = 0) then Lib:= LoadLibrary('D3DX9d.dll');   // MS Debug library - installed with pre-"Feb-05" DirectX9 debug bits
  if (Lib = 0) then Lib:= LoadLibrary('d3dx9_26.dll'); // xxx 2005 SDK - not tested...
  if (Lib = 0) then Exit;

  D3DXAssembleShader:= GetProcAddress(Lib, 'D3DXAssembleShader');
  if (@D3DXAssembleShader = nil) then Exit;

  Result:= True; // Function call is now SUCCEED

  try
    Errors:= '';
    DAShader:= TMemoryStream.Create;
    try
      SaveToStream(DAShader);
      DAShader.Position:= 0;

      Correct:= SUCCEEDED(D3DXAssembleShader(DAShader.Memory, DAShader.Size, nil, nil, 0, @ShaBuff, @ErrBuff));
      if not Correct{ or (ShaBuff = nil)} then
      begin
        Correct:= False;
        SetLength(Errors, ErrBuff.GetBufferSize);
        Move(ErrBuff.GetBufferPointer^, Errors[1], ErrBuff.GetBufferSize);
        Exit;
      end;

      Correct:= (DWORD(FBufferSize) = ShaBuff.GetBufferSize) and
                CompareMem(FBuffer, ShaBuff.GetBufferPointer, ShaBuff.GetBufferSize);
      if not Correct then
      begin
        Errors:= 'Clootie DXShaderDAsm Error: Disassembled Shader compiled to different code!'#10 +
                 'Check wasn''t shader assembed with debug info? Also this may be caused by '#10 +
                 'changed constant values - try to enable high precision contant extraction.';
        Exit;
      end;
    finally
      DAShader.Free;
    end;
  finally
    ShaBuff:= nil;
    ErrBuff:= nil;
    FreeLibrary(Lib);
  end;
end;

end.
